from __future__ import annotations

import shutil
import sys
from typing import Any, Dict, Literal, Tuple

from tyro.constructors._registry import check_default_instances_context
from tyro.constructors._struct_spec import (
    InvalidDefaultInstanceError,
    UnsupportedStructTypeMessage,
)

from . import _fields, _settings, _singleton
from . import _fmtlib as fmt
from .conf import _confstruct, _markers


def _count_matching_fields(
    default: Any,
    subcommand_default: Any,
    subcommand_type: Any,
) -> Tuple[int, int]:
    """Count matching fields between default and subcommand_default.

    Returns (name_matches, value_matches) for lexicographic comparison:
    - name_matches: number of argument names that appear in both
    - value_matches: number of arguments with matching default values

    Note: Type compatibility is already confirmed by _recursive_struct_match.
    This comparison handles edge cases like dictionaries where structural
    comparison via ParserSpecification is more robust than field-level checks.
    """
    if _singleton.is_missing(subcommand_default):
        return (0, 0)

    # Import here to avoid circular dependency.
    from . import _parsers

    def get_arg_defaults(instance: Any, typ: Any) -> Dict[str, Any]:
        """Generate a ParserSpecification and extract {name: default} dict."""
        spec = _parsers.ParserSpecification.from_callable_or_type(
            f=typ,
            markers={_markers.AvoidSubcommands},
            description=None,
            parent_classes=set(),
            default_instance=instance,
            intern_prefix="",
            extern_prefix="",
            subcommand_prefix="",
            support_single_arg_types=True,
            prog_suffix="",
        )
        result: Dict[str, Any] = {}
        for arg_ctx in spec.get_args_including_children():
            arg = arg_ctx.arg
            key = arg.get_output_key()
            result[key] = arg.field.default
        return result

    default_args = get_arg_defaults(default, subcommand_type)
    subcommand_args = get_arg_defaults(subcommand_default, subcommand_type)

    # Count matching names (structural similarity).
    common_names = set(default_args.keys()) & set(subcommand_args.keys())
    name_matches = len(common_names)

    # Count matching values among common names.
    value_matches = 0
    for name in common_names:
        default_val = default_args[name]
        subcommand_val = subcommand_args[name]
        # Ignore any equality check that raises an exception, like numpy arrays.
        try:
            if default_val == subcommand_val:
                value_matches += 1
        except Exception:
            pass

    return (name_matches, value_matches)


def match_subcommand(
    default: Any,
    subcommand_config_from_name: Dict[str, _confstruct._SubcommandConfig],
    subcommand_type_from_name: Dict[str, type],
    extern_prefix: str,
) -> str:
    """Given a subcommand mapping and a default, return which subcommand the default
    corresponds to.

    TOOD: this function is based on heuristics. While it should be robust to
    most real-world scenarios, there's room for improvement for generic types.
    """

    # Get default subcommand name: by identity.
    for subcommand_name, conf in subcommand_config_from_name.items():
        if _singleton.is_missing(conf.default):
            continue
        if default is conf.default:
            return subcommand_name

    # Get default subcommand name: by default value.
    for subcommand_name, conf in subcommand_config_from_name.items():
        if _singleton.is_missing(conf.default):
            continue
        # Ignore any equality check that raises an exception, like numpy arrays.
        try:
            if default == conf.default:
                return subcommand_name
        except Exception:
            continue

    # Find all type-compatible subcommands.
    compatible_matches: list[str] = []
    errors: list[InvalidDefaultInstanceError] = []
    for subcommand_name, subcommand_type in subcommand_type_from_name.items():
        maybe_error = _recursive_struct_match(subcommand_type, default, root=True)
        if not isinstance(maybe_error, InvalidDefaultInstanceError):
            compatible_matches.append(subcommand_name)
        else:
            errors.append(maybe_error)

    # Fast path: if only one match, return it immediately.
    # For argparse backend, always compute similarity to catch errors early.
    use_argparse_backend = _settings._experimental_options["backend"] == "argparse"
    if len(compatible_matches) == 1 and not use_argparse_backend:
        return compatible_matches[0]

    # Multiple matches (or argparse backend): use field counting to pick the best.
    if len(compatible_matches) > 0:
        best_match: str | None = None
        best_count: Tuple[int, int] = (-1, -1)
        for subcommand_name in compatible_matches:
            subcommand_type = subcommand_type_from_name[subcommand_name]
            conf = subcommand_config_from_name.get(subcommand_name)
            conf_default = (
                conf.default if conf is not None else _singleton.MISSING_NONPROP
            )
            count = _count_matching_fields(default, conf_default, subcommand_type)
            if count > best_count:
                best_count = count
                best_match = subcommand_name

        assert best_match is not None
        return best_match

    # No compatible matches found. Print a detailed error message showing why each
    # subcommand was rejected.
    details = []
    for subcommand_name, error in zip(subcommand_type_from_name, errors):
        details.append("")
        details.append(
            fmt.text(
                fmt.text["yellow", "bold"](subcommand_name),
                " was not a match because:",
            )
        )
        # Add each message in the tuple as its own bullet point.
        for msg in error.message:
            details.append(fmt.cols((fmt.text["dim"]("• "), 2), msg))
    details.append("")
    details.append(fmt.hr["red"]())
    details.append(
        "Debugging: check that the field default matches a member of the union type."
    )

    error_message = fmt.box["bright_red"](
        fmt.text["bright_red", "bold"]("Invalid input to tyro.cli()"),
        fmt.rows(
            fmt.text(
                "The default value of the ",
                fmt.text["green", "bold"](extern_prefix),
                " field could not be matched to any of its subcommands.",
            ),
            fmt.hr["red"](),
            *details,
        ),
    )
    print(
        "\n".join(error_message.render(width=min(shutil.get_terminal_size()[0], 80))),
        file=sys.stderr,
        flush=True,
    )
    sys.exit(2)


def _recursive_struct_match(
    subcommand_type: Any, default: Any, root: bool, intern_prefix: str = ""
) -> Literal[True] | InvalidDefaultInstanceError:
    """Returns `True` if the given type and default instance are compatible
    with each other."""
    # Can we generate a field list from this type?
    with check_default_instances_context():
        maybe_field_list = _fields.field_list_from_type_or_callable(
            subcommand_type,
            default,
            support_single_arg_types=root,
            in_union_context=False,
        )

    # Found a struct type that matches, but the default instance isn't compatible.
    if isinstance(maybe_field_list, InvalidDefaultInstanceError):
        return maybe_field_list

    # Base case: found a leaf.
    if isinstance(maybe_field_list, UnsupportedStructTypeMessage):
        return True

    field_list = maybe_field_list
    for field in field_list[1]:
        field_check = _recursive_struct_match(
            field.type,
            field.default,
            root=False,
            intern_prefix=intern_prefix + field.intern_name + ".",
        )
        if isinstance(field_check, InvalidDefaultInstanceError):
            # Add context about which field failed.
            field_path = (
                field.intern_name
                if intern_prefix == ""
                else intern_prefix + field.intern_name
            )
            return InvalidDefaultInstanceError(
                (
                    fmt.text(
                        "Field ",
                        fmt.text["magenta", "bold"](field_path),
                        " has invalid default",
                    ),
                    *field_check.message,
                )
            )

    return True



"""Subcommands as function arguments

A subcommand will be created for each input annotated with a union over
struct types.

.. note::

    To prevent :func:`tyro.cli()` from converting a Union type into a subcommand,
    use :class:`tyro.conf.AvoidSubcommands`.

.. note::

    Argument ordering for subcommands can be tricky. In the example below,
    ``--shared-arg`` must always come *before* the subcommand. As an option for
    alleviating this, see :class:`tyro.conf.CascadeSubcommandArgs`.


Usage:

    # Print the helptext. This will show the available subcommands:
    python ./02_subcommands_in_func.py --help

    # Using the default subcommand:
    python ./02_subcommands_in_func.py --shared-arg 100

    # Choosing a different subcommand:
    python ./02_subcommands_in_func.py --shared-arg 100 cmd:commit --cmd.message Hello!
"""

from __future__ import annotations

import dataclasses

import tyro


@dataclasses.dataclass
class Checkout:
    """Checkout a branch."""

    branch: str


@dataclasses.dataclass
class Commit:
    """Commit changes."""

    message: str


def main(
    shared_arg: int,
    cmd: Checkout | Commit = Checkout(branch="default"),
):
    print(f"{shared_arg=}")
    print(cmd)


if __name__ == "__main__":
    tyro.cli(main)



"""Rules for taking high-level field definitions and lowering them into inputs for
argparse's `add_argument()`."""

from __future__ import annotations

import collections.abc
import dataclasses
import json
import shlex
from functools import cached_property
from typing import (
    Any,
    Callable,
    Iterable,
    Literal,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import get_args, get_origin

from . import _fields, _settings, _singleton, _strings
from . import _fmtlib as fmt
from ._backends import _argparse as argparse
from ._typing_compat import is_typing_union
from .conf import _markers
from .constructors import (
    ConstructorRegistry,
    PrimitiveTypeInfo,
    UnsupportedTypeAnnotationError,
)

_T = TypeVar("_T")


def flag_to_inverse(option_string: str) -> str | None:
    """Converts --flag to --no-flag, --child.flag to --child.no-flag, etc.

    Returns None for short flags (those not starting with '--'), since
    single-letter aliases like -f cannot have a meaningful inverse form.
    """
    # Short flags (like -f) cannot be inverted.
    if not option_string.startswith("--"):
        return None
    if "." not in option_string:
        option_string = "--no" + _strings.get_delimeter() + option_string[2:]
    else:
        # Loose heuristic for where to add the no-/no_ prefix.
        left, _, right = option_string.rpartition(".")
        option_string = left + ".no" + _strings.get_delimeter() + right
    return option_string


class BooleanOptionalAction(argparse.Action):
    """Adapted from https://github.com/python/cpython/pull/27672"""

    def __init__(
        self,
        option_strings: Sequence[str],
        dest: str,
        default: _T | str | None = None,
        type: Callable[[str], _T] | argparse.FileType | None = None,
        choices: Iterable[_T] | None = None,
        required: bool = False,
        help: str | None = None,
        metavar: str | tuple[str, ...] | None = None,
    ) -> None:
        _option_strings = []
        self._no_strings = set()
        for option_string in option_strings:
            _option_strings.append(option_string)

            if option_string.startswith("--"):
                option_string = flag_to_inverse(option_string)
                self._no_strings.add(option_string)
                _option_strings.append(option_string)

        super().__init__(
            option_strings=_option_strings,
            dest=dest,
            nargs=0,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        if option_string in self.option_strings:
            assert option_string is not None
            setattr(namespace, self.dest, option_string not in self._no_strings)


@dataclasses.dataclass(frozen=True)
class ArgumentDefinition:
    """Structure containing everything needed to define an argument."""

    intern_prefix: str  # True prefix. (eg for the argument's dest field)
    extern_prefix: str  # User-facing prefix.
    subcommand_prefix: str  # Prefix for nesting.
    field: _fields.FieldDefinition

    def get_output_key(self) -> str:
        """Get key used for this arg in the parsed output dict."""
        if self.is_positional():
            return self.lowered.name_or_flags[-1]
        else:
            assert self.lowered.dest is not None
            return self.lowered.dest

    def is_positional(self) -> bool:
        """Returns True if the argument should be positional in the commandline."""
        return (
            # Explicit positionals.
            _markers.Positional in self.field.markers
            or (
                # Make required arguments positional.
                _markers.PositionalRequiredArgs in self.field.markers
                and _singleton.is_missing(self.field.default)
            )
            # Argumens with no names (like in DummyWrapper) should be
            # positional.
            or (
                self.field.extern_name == ""
                and self.field.intern_name == "__tyro_dummy_inner__"
            )
        )

    def add_argument(
        self, parser: Union[argparse.ArgumentParser, argparse._ArgumentGroup]
    ) -> None:
        """Add a defined argument to a parser."""

        # Get keyword arguments, with None values removed.
        kwargs = dict(self.lowered.__dict__)  # type: ignore
        kwargs.pop("instance_from_str")
        kwargs.pop("str_from_instance")
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        name_or_flags = kwargs.pop("name_or_flags")

        # Evaluate lazy help if callable.
        if "help" in kwargs and callable(kwargs["help"]):
            kwargs["help"] = kwargs["help"]()

        if self.is_positional():
            if "required" in kwargs:
                kwargs.pop("required")  # Can't be passed in for positional arguments.
            if len(name_or_flags) > 1:
                import warnings

                warnings.warn(
                    f"Aliases were specified, but {name_or_flags} is positional. Aliases will be ignored."
                )
                name_or_flags = name_or_flags[-1:]

        if kwargs.get("action", None) == "boolean_optional_action":
            kwargs["action"] = BooleanOptionalAction

        # Add argument, with aliases if available.
        arg = parser.add_argument(*name_or_flags, **kwargs)

        # Do our best to tab complete paths.
        # There will be false positives here, but if choices is unset they should be
        # harmless.
        # Note: shtab is now optional, so we only set completion hints if available.
        if "choices" not in kwargs:
            name_suggests_dir = (
                # The conditions are intended to be conservative; if a directory path is
                # registered as a normal file one that's OK, the reverse on the other
                # hand will be overly restrictive.
                self.field.intern_name.endswith("_dir")
                or self.field.intern_name.endswith("_directory")
                or self.field.intern_name.endswith("_folder")
            )
            name_suggests_path = (
                self.field.intern_name.endswith("_file")
                or self.field.intern_name.endswith("_path")
                or self.field.intern_name.endswith("_filename")
                or name_suggests_dir
            )
            complete_as_path = (
                # Catch types like Path, List[Path], Tuple[Path, ...] etc.
                "Path" in str(self.field.type_stripped)
                # For string types, we require more evidence.
                or ("str" in str(self.field.type_stripped) and name_suggests_path)
            )
            if complete_as_path:
                try:
                    import shtab

                    arg.complete = shtab.DIRECTORY if name_suggests_dir else shtab.FILE  # type: ignore
                except ImportError:
                    # shtab is optional; if not available, skip completion hints.
                    pass

    @cached_property
    def lowered(self) -> LoweredArgumentDefinition:
        """Lowered argument definition, generated by applying a sequence of rules."""
        # Each rule will mutate the lowered object. This is (unfortunately)
        # much faster than a functional approach.
        lowered = LoweredArgumentDefinition()
        _rule_handle_boolean_flags(self, lowered)
        _rule_apply_primitive_specs(self, lowered)
        _rule_counters(self, lowered)
        _rule_generate_helptext(self, lowered)
        _rule_set_name_or_flag_and_dest(self, lowered)
        _rule_positional_special_handling(self, lowered)
        _rule_apply_argconf(self, lowered)
        return lowered

    def is_suppressed(self) -> bool:
        """Returns if the argument is suppressed. Suppressed arguments won't be
        added to the parser."""
        return _markers.Suppress in self.field.markers or (
            _markers.SuppressFixed in self.field.markers and self.lowered.is_fixed()
        )

    def get_invocation_text(self) -> tuple[fmt._Text, fmt._Text]:
        """Returns (invocation short, invocation long)."""

        # Use bright red for fixed arguments, otherwise just bold.
        metavar_style = ("bright_red", "bold") if self.lowered.is_fixed() else ("bold",)
        if self.is_positional():
            assert self.lowered.metavar is not None
            invocation_short = fmt.text[metavar_style](self.lowered.metavar)
            invocation_long = fmt.text(self.lowered.metavar)
            return invocation_short, invocation_long

        name_or_flags: list[str] = list(self.lowered.name_or_flags)
        if self.lowered.action == "boolean_optional_action":
            name_or_flags = []
            for name_or_flag in self.lowered.name_or_flags:
                name_or_flags.append(name_or_flag)
                # Short flags (like -f) cannot be inverted.
                inv = flag_to_inverse(name_or_flag)
                if inv is not None:
                    name_or_flags.append(inv)
            # Find the first invertible flag for the short invocation display.
            first_inv = None
            for name_or_flag in self.lowered.name_or_flags:
                first_inv = flag_to_inverse(name_or_flag)
                if first_inv is not None:
                    break
            # There should always be at least one long flag (starting with --) because:
            # - The main argument name always gets -- prepended (see _rule_generate_name_or_flag)
            # - tyro.conf.arg(name=...) prepends -- to the custom name
            # - tyro.conf.arg(aliases=...) only adds flags, never removes the main one
            assert first_inv is not None
            invocation_short = fmt.text(
                self.lowered.name_or_flags[0],
                " | ",
                first_inv,
            )
        elif self.lowered.metavar is not None:
            invocation_short = fmt.text(
                self.lowered.name_or_flags[0],
                " ",
                fmt.text[metavar_style](self.lowered.metavar),
            )
        else:
            invocation_short = fmt.text(self.lowered.name_or_flags[0])

        if self.lowered.required is not True:
            invocation_short = fmt.text("[", invocation_short, "]")

        invocation_long_parts: list[str | fmt._Text] = []
        for i, name in enumerate(name_or_flags):
            if i > 0:
                invocation_long_parts.append(", ")

            invocation_long_parts.append(name)
            if self.lowered.metavar is not None and self.lowered.metavar != "":
                invocation_long_parts.append(" ")
                invocation_long_parts.append(
                    fmt.text[metavar_style](self.lowered.metavar)
                )

        return invocation_short, fmt.text(*invocation_long_parts)


@dataclasses.dataclass
class LoweredArgumentDefinition:
    """Contains fields meant to be passed directly into argparse."""

    # Action that is called on parsed arguments. This handles conversions from strings
    # to our desired types.
    #
    # The main reason we use this instead of the standard 'type' argument is to enable
    # mixed-type tuples.
    instance_from_str: Optional[Callable] = None
    str_from_instance: Optional[Callable] = None

    def is_fixed(self) -> bool:
        """If the instantiator is set to `None`, even after all argument
        transformations, it means that we don't have a valid instantiator for an
        argument. We then mark the argument as 'fixed', with a value always equal to the
        field default."""
        return self.instance_from_str is None

    # From here on out, all fields correspond 1:1 to inputs to argparse's
    # add_argument() method.
    name_or_flags: Tuple[str, ...] = ()
    default: Optional[Any] = None
    dest: Optional[str] = None
    required: Optional[bool] = None
    action: Optional[
        Literal[
            "count", "append", "store_true", "store_false", "boolean_optional_action"
        ]
    ] = None
    nargs: Optional[Union[int, Literal["*", "?"]]] = None
    choices: Optional[Tuple[str, ...]] = None
    # Note: unlike in vanilla argparse, our metavar is always a string. We handle
    # sequences, multiple arguments, etc, manually.
    metavar: Optional[str] = None
    help: Optional[str] | Callable[[], str] = None


def _get_single_non_none_union_arg(typ: Any) -> Optional[Any]:
    """If typ is Union[T, None], return T. Otherwise return None."""
    if not is_typing_union(get_origin(typ)):
        return None
    type_args = get_args(typ)
    non_none_args = [t for t in type_args if t is not type(None)]
    if len(non_none_args) == 1:
        return non_none_args[0]
    return None


def _rule_handle_boolean_flags(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> None:
    # Check if this is a plain bool type.
    is_bool = arg.field.type_stripped is bool

    # Check if this is DisallowNone[bool | None], which should behave like bool
    # from a CLI perspective since the user can only choose True or False.
    if not is_bool and _markers.DisallowNone in arg.field.markers:
        inner_type = _get_single_non_none_union_arg(arg.field.type_stripped)
        is_bool = inner_type is bool

    if not is_bool:
        return

    if (
        _singleton.is_missing(arg.field.default)
        or arg.is_positional()
        or _markers.FlagConversionOff in arg.field.markers
        or _markers.Fixed in arg.field.markers
    ):
        # Treat bools as a normal parameter.
        return

    # Default `False` => --flag passed in flips to `True`.
    if _markers.FlagCreatePairsOff in arg.field.markers:
        # If default is True, --flag will flip to `False`.
        # If default is False, --no-flag will flip to `True`.
        lowered.action = "store_false" if arg.field.default else "store_true"
    else:
        # Create both --flag and --no-flag.
        lowered.action = "boolean_optional_action"
    lowered.instance_from_str = lambda x: x  # argparse will directly give us a bool!
    lowered.default = arg.field.default
    return


def _rule_apply_primitive_specs(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> None:
    """The bulkiest bit: recursively analyze the type annotation and use it to determine
    how to instantiate it given some string from the commandline.

    Important: as far as argparse is concerned, all inputs are strings.

    Conversions from strings to our desired types happen in the instantiator; this is a
    bit more flexible, and lets us handle more complex types like enums and multi-type
    tuples."""

    if _markers.Fixed in arg.field.markers:
        lowered.instance_from_str = None
        lowered.metavar = "{fixed}"
        lowered.required = False
        lowered.default = _singleton.MISSING
        return
    if lowered.instance_from_str is not None:
        lowered.required = False
        return

    # For positional arguments, UseAppendAction should be ignored since you can't
    # repeat a positional like `--flag val1 --flag val2`. Generate spec without it.
    exclude_markers = (
        {_markers.UseAppendAction}
        if arg.is_positional() and _markers.UseAppendAction in arg.field.markers
        else None
    )

    spec = ConstructorRegistry.get_primitive_spec(
        PrimitiveTypeInfo.make(
            cast(type, arg.field.type),
            arg.field.markers,
            exclude_markers=exclude_markers,
        )
    )
    if isinstance(spec, UnsupportedTypeAnnotationError):
        error = spec
        if _singleton.is_missing(arg.field.default):
            field_name = _strings.make_field_name(
                [arg.extern_prefix, arg.field.extern_name]
            )
            if field_name != "":
                raise UnsupportedTypeAnnotationError(
                    (
                        fmt.text(
                            "Unsupported type annotation for field ",
                            fmt.text["magenta", "bold"](field_name),
                            " with type ",
                            fmt.text["cyan"](str(arg.field.type)),
                        ),
                        *error.message,
                        fmt.text(
                            "To suppress this error, assign the field either a default value or a different type"
                        ),
                    )
                )
            else:
                # If the field name is empty, it means we're raising an error
                # for the direct input to `tyro.cli()`. We don't need to write
                # out which specific field we're complaining about.
                raise UnsupportedTypeAnnotationError(
                    (
                        *error.message,
                        fmt.text(
                            "To suppress this error, assign the field either a default value or a different type"
                        ),
                    )
                )
        else:
            # For fields with a default, we'll get by even if there's no instantiator
            # available.
            lowered.metavar = "{fixed}"
            lowered.required = False
            lowered.default = _singleton.MISSING
            return

    # Mark lowered as required if a default is missing.
    if (
        _singleton.is_missing(arg.field.default)
        and _markers._OPTIONAL_GROUP not in arg.field.markers
    ):
        lowered.required = True

    # We're actually going to skip the default field: if an argument is unset, the
    # MISSING value will be detected in _calling.py and the field default will
    # directly be used. This helps reduce the likelihood of issues with converting
    # the field default to a string format, then back to the desired type.
    if spec._action == "append":
        lowered.default = []
    else:
        lowered.default = _singleton.MISSING_NONPROP

    if spec._action == "append":

        def append_instantiator(x: list[list[str]]) -> Any:
            """Handle UseAppendAction effects."""
            # We'll assume that the type is annotated as Dict[...], Tuple[...], List[...], etc.
            container_type = get_origin(arg.field.type_stripped)
            if container_type is None:
                # Raw annotation, like `UseAppendAction[list]`. It's unlikely
                # that a user would use this but we can handle it.
                container_type = arg.field.type_stripped

            # Instantiate initial output.
            out = (
                arg.field.default
                if not _singleton.is_missing(arg.field.default)
                else None
            )
            if out is None:
                out = {} if container_type is dict else []
            elif isinstance(out, dict):
                out = out.copy()
            else:
                # All sequence types will be lists for now to make sure we can
                # append to them.
                out = list(out)

            # Get + merge parts.
            parts = [spec.instance_from_str(arg_list) for arg_list in x]
            for part in parts:
                if isinstance(out, dict):
                    out.update(part)
                else:
                    out.append(part)

            # Return output with correct type.
            if container_type in (dict, Sequence, collections.abc.Sequence):
                return out
            else:
                return container_type(out)

        lowered.instance_from_str = append_instantiator
        lowered.str_from_instance = spec.str_from_instance
        lowered.choices = spec.choices
        lowered.nargs = spec.nargs if not isinstance(spec.nargs, tuple) else "*"
        lowered.metavar = spec.metavar
        lowered.action = spec._action
        lowered.required = False
        return
    else:
        lowered.instance_from_str = spec.instance_from_str
        lowered.str_from_instance = spec.str_from_instance
        lowered.choices = spec.choices
        lowered.nargs = spec.nargs if not isinstance(spec.nargs, tuple) else "*"
        lowered.metavar = spec.metavar
        lowered.action = spec._action
        return


def _rule_counters(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> None:
    """Handle counters, like -vvv for level-3 verbosity."""
    if (
        _markers.UseCounterAction in arg.field.markers
        and arg.field.type_stripped is int
        and not arg.is_positional()
    ):
        lowered.metavar = None
        lowered.nargs = None
        lowered.action = "count"
        lowered.default = 0
        lowered.required = False
        lowered.instance_from_str = (
            lambda x: x
        )  # argparse will directly give us an int!
        return


def _rule_generate_helptext(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> None:
    """Store context for lazy helptext generation."""
    # Defer helptext generation by wrapping in a callable.
    # The percent symbol needs some extra handling in argparse.
    # https://stackoverflow.com/questions/21168120/python-argparse-errors-with-in-help-string
    lowered.help = lambda: (
        generate_argument_helptext(arg, lowered).as_str_no_ansi().replace("%", "%%")
    )


def _rule_set_name_or_flag_and_dest(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> None:
    extern_name = arg.field.extern_name
    if lowered.action == "store_false":
        extern_name = "no_" + extern_name

    if (
        arg.field.argconf.prefix_name is False
        or _markers.OmitArgPrefixes in arg.field.markers
    ):
        # Strip prefixes when the argument is suppressed.
        # Still need to call make_field_name() because it converts underscores
        # to hyphens, etc.
        name_or_flag = _strings.make_field_name([extern_name])
    elif (
        _markers.OmitSubcommandPrefixes in arg.field.markers
        and arg.subcommand_prefix != ""
    ):
        # Strip subcommand prefixes, but keep following
        # prefixes.`extern_prefix` can start with the prefix corresponding to
        # the parent subcommand, but end with other prefixes correspondeding to
        # nested structures within the subcommand.
        name_or_flag = _strings.make_field_name([arg.extern_prefix, extern_name])
        strip_prefix = arg.subcommand_prefix + "."
        assert name_or_flag.startswith(strip_prefix), name_or_flag
        name_or_flag = name_or_flag[len(strip_prefix) :]
    else:
        # Standard prefixed name.
        name_or_flag = _strings.make_field_name([arg.extern_prefix, extern_name])

    # Prefix keyword arguments with --.
    if not arg.is_positional():
        name_or_flag = "--" + name_or_flag

    lowered.name_or_flags = (name_or_flag,)
    lowered.dest = _strings.make_field_name([arg.intern_prefix, arg.field.intern_name])


def _rule_positional_special_handling(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> None:
    if not arg.is_positional():
        return None

    metavar = lowered.metavar
    if lowered.required:
        nargs = lowered.nargs
    else:
        if metavar is not None:
            metavar = "[" + metavar + "]"
        if lowered.nargs == 1:
            # Optional positional arguments. This needs to be special-cased in
            # _calling.py.
            nargs = "?"
        else:
            # If lowered.nargs is either + or an int.
            nargs = "*"

    lowered.name_or_flags = (
        _strings.make_field_name([arg.intern_prefix, arg.field.intern_name]),
    )
    lowered.dest = None
    lowered.metavar = metavar
    lowered.nargs = nargs
    return


def _rule_apply_argconf(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> None:
    if arg.field.argconf.metavar is not None:
        lowered.metavar = arg.field.argconf.metavar
    if arg.field.argconf.aliases is not None:
        lowered.name_or_flags = arg.field.argconf.aliases + lowered.name_or_flags


def generate_argument_helptext(
    arg: ArgumentDefinition, lowered: LoweredArgumentDefinition, compact: bool = False
) -> fmt._Text:
    help_parts: list[str | fmt._Text] = []

    primary_help = arg.field.helptext
    # Evaluate lazy helptext if it's a callable.
    if callable(primary_help):
        primary_help = primary_help()

    if primary_help is None and _markers.Positional in arg.field.markers:
        primary_help = _strings.make_field_name(
            [arg.extern_prefix, arg.field.intern_name]
        )

    # In compact mode, skip the primary help text.
    if primary_help is not None and not compact:
        help_parts.append(fmt.text["dim"](primary_help))

    if not lowered.required:
        # Get the default value.
        # Note: lowered.default is the stringified version!
        if (
            lowered.is_fixed()
            or lowered.action
            in (
                "append",
                "boolean_optional_action",
                "store_true",
                "store_false",
            )
            or _singleton.is_missing(arg.field.default)
        ):
            # Cases where we want to use the field default directly.
            default = arg.field.default
        elif arg.field.default is _singleton.EXCLUDE_FROM_CALL:
            default = None
        else:
            # Standard cases where we convert to a string representation.
            default = (
                lowered.str_from_instance(arg.field.default)
                if lowered.str_from_instance is not None
                else None
            )

        # Get the default value label.
        if arg.field.argconf.constructor_factory is not None:
            default_label = (
                str(default)
                if arg.field.type_stripped is not json.loads
                else json.dumps(arg.field.default)
            )
        elif type(default) in (tuple, list, set):
            # For tuple types, we might have default as (0, 1, 2, 3).
            # For list types, we might have default as [0, 1, 2, 3].
            # For set types, we might have default as {0, 1, 2, 3}.
            #
            # In all cases, we want to display (default: 0 1 2 3), for consistency with
            # the format that argparse expects when we set nargs.
            assert default is not None
            default_label = " ".join(map(shlex.quote, map(str, default)))
        else:
            default_label = str(default)

        # Suffix helptext with some behavior hint, such as the default value of the argument.
        help_behavior_hint = arg.field.argconf.help_behavior_hint
        if help_behavior_hint is not None:
            behavior_hint = (
                help_behavior_hint(default_label)
                if callable(help_behavior_hint)
                else help_behavior_hint
            )
        elif lowered.instance_from_str is None:
            # Intentionally not quoted via shlex, since this can't actually be passed
            # in via the commandline.
            behavior_hint = f"(fixed to: {str(default)})"
        elif lowered.action == "count":
            # Repeatable argument.
            behavior_hint = "(repeatable)"
        elif lowered.action == "append" and (
            _singleton.is_missing(default) or len(cast(tuple, default)) == 0
        ):
            behavior_hint = "(repeatable)"
        elif lowered.action == "append" and len(cast(tuple, default)) > 0:
            assert default is not None  # Just for type checker.
            behavior_hint = f"(repeatable, appends to: {default_label})"
        elif arg.field.default is _singleton.EXCLUDE_FROM_CALL:
            # ^important to use arg.field.default and not the stringified default variable.
            behavior_hint = "(unset by default)"
        elif _markers._OPTIONAL_GROUP in arg.field.markers and _singleton.is_missing(
            default
        ):
            # Argument in an optional group, but with no default. This is typically used
            # when general (non-argument, non-dataclass) object arguments are given a
            # default, or when we use `tyro.conf.arg(constructor=...)`.
            #
            # There are some usage details that aren't communicated right now in the
            # helptext. For example: all arguments within an optional group without a
            # default should be passed in or none at all.
            behavior_hint = "(optional)"
        elif _markers._OPTIONAL_GROUP in arg.field.markers:
            # Argument in an optional group, but which also has a default.
            behavior_hint = f"(default if used: {default_label})"
        else:
            behavior_hint = f"(default: {default_label})"

        help_parts.append(
            fmt.text[
                _settings.ACCENT_COLOR if _settings.ACCENT_COLOR != "white" else "cyan"
            ](behavior_hint)
        )
    else:
        help_parts.append(fmt.text["bright_red"]("(required)"))

    return fmt.text(*help_parts, delimeter=" ")



"""Generics (Python <3.12)

The legacy :py:class:`typing.Generic` and :py:class:`typing.TypeVar` syntax for
generic types is also supported.

Usage:

    python ./02_generics.py --help
"""

import dataclasses
from typing import Generic, TypeVar

import tyro

ScalarType = TypeVar("ScalarType", int, float)
ShapeType = TypeVar("ShapeType")


@dataclasses.dataclass
class Point3(Generic[ScalarType]):
    x: ScalarType
    y: ScalarType
    z: ScalarType
    frame_id: str


@dataclasses.dataclass
class Triangle:
    a: Point3[float]
    b: Point3[float]
    c: Point3[float]


@dataclasses.dataclass
class Args(Generic[ShapeType]):
    shape: ShapeType


if __name__ == "__main__":
    args = tyro.cli(Args[Triangle])
    print(args)



"""Compact help

When a configuration has many fields, the help text can become verbose and
difficult to scan. The ``compact_help`` parameter in :func:`tyro.cli()` enables
a more succinct format that omits field descriptions.

In compact mode, users can still access full descriptions via ``--help-verbose``.

Usage:
    # Compact mode (default with compact_help=True):
    python ./15_compact_help.py --help

    # Verbose mode (access full descriptions):
    python ./15_compact_help.py --help-verbose

    # Actually run it:
    python ./15_compact_help.py --host localhost --port 8080
"""

from dataclasses import dataclass
from typing import Literal

import tyro


@dataclass
class ServerConfig:
    """Configuration for a web server."""

    # Network settings.
    host: str = "0.0.0.0"
    """Hostname to bind the server to."""

    port: int = 8000
    """Port number to listen on."""

    workers: int = 4
    """Number of worker processes."""

    timeout: int = 30
    """Request timeout in seconds."""

    # Security settings.
    ssl_enabled: bool = False
    """Enable SSL/TLS encryption."""

    ssl_cert_path: str = "/etc/ssl/cert.pem"
    """Path to SSL certificate file."""

    ssl_key_path: str = "/etc/ssl/key.pem"
    """Path to SSL private key file."""

    # Application settings.
    max_request_size: int = 10485760
    """Maximum request size in bytes."""

    cors_origins: str = "*"
    """Allowed CORS origins (comma-separated)."""

    log_level: Literal["debug", "info", "warning", "error"] = "info"
    """Logging level for the application."""

    log_file: str = "/var/log/server.log"
    """Path to log file."""

    # Performance settings.
    cache_enabled: bool = True
    """Enable response caching."""

    cache_size: int = 1000
    """Maximum number of cached responses."""

    compression_enabled: bool = True
    """Enable response compression."""

    keepalive_timeout: int = 5
    """Keep-alive connection timeout in seconds."""

    # Database settings.
    db_host: str = "localhost"
    """Database server hostname."""

    db_port: int = 5432
    """Database server port."""

    db_name: str = "appdb"
    """Database name."""


if __name__ == "__main__":
    # Parse with compact_help=True to enable compact mode.
    config = tyro.cli(ServerConfig, compact_help=True)
    print(f"Starting server with config:\n{config}")



from __future__ import annotations

import collections
import collections.abc
import dataclasses
import datetime
import enum
import inspect
import itertools
import json
import os
import pathlib
import sys
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Literal,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import TYPE_CHECKING, assert_never, get_args, get_origin

from .._typing_compat import is_typing_literal, is_typing_union

if TYPE_CHECKING:
    from ._registry import ConstructorRegistry

from .. import _fmtlib as fmt
from .. import _resolver, _strings
from .._typing import TypeForm
from ..conf import _markers
from ._backtracking import parse_with_backtracking


class UnsupportedTypeAnnotationError(Exception):
    """Exception raised when a type annotation is not supported."""

    def __init__(self, message: tuple[fmt._Text, ...]):
        self.message = message
        super().__init__()


T = TypeVar("T")


@dataclasses.dataclass(frozen=True)
class PrimitiveTypeInfo:
    """Information used to generate constructors for primitive types."""

    type: TypeForm
    """Annotated field type. Forward references, aliases, and type
    variables/parameters will have been resolved and runtime annotations
    (typing.Annotated) will have been stripped."""
    type_origin: TypeForm | None
    """The output of get_origin() on the static type."""
    markers: set[_markers.Marker]
    """Set of tyro markers used to configure this field."""
    _primitive_spec: PrimitiveConstructorSpec | None
    """Primitive constructor spec that was scraped from runtime annotations."""

    @staticmethod
    def make(
        raw_annotation: TypeForm | Callable,
        parent_markers: set[_markers.Marker],
        exclude_markers: set[_markers.Marker] | None = None,
    ) -> PrimitiveTypeInfo:
        _, primitive_specs = _resolver.unwrap_annotated(
            raw_annotation, search_type=PrimitiveConstructorSpec
        )
        primitive_spec = primitive_specs[0] if len(primitive_specs) > 0 else None

        typ, extra_markers = _resolver.unwrap_annotated(
            raw_annotation, search_type=_markers._Marker
        )
        markers = parent_markers | set(extra_markers)
        if exclude_markers is not None:
            markers = markers - exclude_markers
        return PrimitiveTypeInfo(
            type=cast(TypeForm, typ),
            type_origin=get_origin(typ),
            markers=markers,
            _primitive_spec=primitive_spec,
        )


@dataclasses.dataclass(frozen=True)
class PrimitiveConstructorSpec(Generic[T]):
    """Specification for constructing a primitive type from a string.

    There are two ways to use this class:

    First, we can include it in a type signature via :class:`typing.Annotated`.
    This is the simplest for making local modifications to parsing behavior for
    individual fields.

    Alternatively, it can be returned by a rule in a :class:`ConstructorRegistry`.
    """

    nargs: int | tuple[int, ...] | Literal["*"]
    """Number of arguments required to construct an instance. If nargs is "*", then
    the number of arguments is variable. If nargs is a tuple, it represents multiple
    valid fixed argument counts (used for unions with different fixed nargs)."""
    metavar: str
    """Metavar to display in help messages."""
    instance_from_str: Callable[[list[str]], T]
    """Given a list of string arguments, construct an instance of the type. The
    length of the list will match the value of nargs."""
    is_instance: Callable[[Any], bool | Literal["~"]]
    """Given an object instance, does it match this primitive type? This is
    used for specific help messages when both a union type is present and a
    default is provided.

    Can return "~" to signify that an instance is a "fuzzy" match, and should
    only be used if there are no other matches. This is used for numeric tower
    support.
    """
    str_from_instance: Callable[[T], list[str]]
    """Convert an instance to a list of string arguments that would construct
    the instance. This is used for help messages when a default is provided."""
    choices: tuple[str, ...] | None = None
    """Finite set of choices for arguments."""

    _action: Literal["append"] | None = None
    """Internal action to use. Not part of the public API."""


def _compute_total_nargs(
    specs: Sequence[PrimitiveConstructorSpec],
) -> int | tuple[int, ...] | Literal["*"]:
    """Compute all possible total argument counts for a sequence of specs.

    When specs have tuple nargs (representing multiple valid argument counts),
    this computes all possible sums.

    Note: This uses itertools.product which is exponential in the number of specs,
    but in practice we're only combining 2-3 specs (e.g., dict keys and values,
    or tuple elements). Could be optimized in the future.

    Args:
        specs: Sequence of PrimitiveConstructorSpec objects.

    Returns:
        Total nargs value: either a single int, a tuple of ints, or "*".
    """
    nargs_options_per_spec: list[tuple[int, ...]] = []
    is_variable_nargs = False

    for spec in specs:
        if spec.nargs == "*":
            is_variable_nargs = True
        elif isinstance(spec.nargs, int):
            nargs_options_per_spec.append((spec.nargs,))
        elif isinstance(spec.nargs, tuple):
            nargs_options_per_spec.append(tuple(sorted(spec.nargs)))
        else:
            assert_never(spec.nargs)

    if is_variable_nargs:
        # If any spec has nargs='*', the total nargs is variable.
        return "*"
    else:
        possible_totals = {
            sum(combo) for combo in itertools.product(*nargs_options_per_spec)
        }
        nargs = tuple(sorted(possible_totals))
        if len(nargs) == 1:
            return nargs[0]
        return nargs


def apply_default_primitive_rules(registry: ConstructorRegistry) -> None:
    """Apply default rules to the registry."""

    from ._registry import ConstructorRegistry

    @registry.primitive_rule
    def any_rule(
        type_info: PrimitiveTypeInfo,
    ) -> PrimitiveConstructorSpec | UnsupportedTypeAnnotationError | None:
        if type_info.type is not Any:
            return None
        return UnsupportedTypeAnnotationError(
            (fmt.text("`Any` is not a parsable type."),)
        )

    # HACK (json.loads): this is for code that uses
    # `tyro.conf.arg(constructor=json.loads)`. We're going to deprecate this
    # syntax (the constructor= argument in tyro.conf.arg), but there is code
    # that lives in the wild that relies on the behavior so we'll do our best
    # not to break it.
    vanilla_types = (int, str, float, complex, bytes, bytearray, json.loads)

    @registry.primitive_rule
    def basics_rule(type_info: PrimitiveTypeInfo) -> PrimitiveConstructorSpec | None:
        if type_info.type not in vanilla_types:
            return None
        return PrimitiveConstructorSpec(
            nargs=1,
            metavar=type_info.type.__name__.upper(),
            instance_from_str=lambda args: (
                bytes(args[0], encoding="ascii")
                if type_info.type is bytes
                else type_info.type(args[0])
            ),
            # issubclass(type(x), y) here is preferable over isinstance(x, y)
            # due to quirks in the numeric tower.
            is_instance=lambda x: _resolver.isinstance_with_fuzzy_numeric_tower(
                x, type_info.type
            ),
            str_from_instance=lambda instance: [str(instance)],
        )

    if "torch" in sys.modules.keys():
        import torch

        @registry.primitive_rule
        def torch_device_rule(
            type_info: PrimitiveTypeInfo,
        ) -> PrimitiveConstructorSpec | None:
            if type_info.type is not torch.device:
                return None
            return PrimitiveConstructorSpec(
                nargs=1,
                metavar=type_info.type.__name__.upper(),
                instance_from_str=lambda args: torch.device(args[0]),
                is_instance=lambda x: isinstance(x, type_info.type),
                str_from_instance=lambda instance: [str(instance)],
            )

    @registry.primitive_rule
    def bool_rule(type_info: PrimitiveTypeInfo) -> PrimitiveConstructorSpec | None:
        if type_info.type is not bool:
            return None
        return PrimitiveConstructorSpec(
            nargs=1,
            metavar="{True,False}",
            instance_from_str=lambda args: args[0] == "True",
            choices=("True", "False"),
            is_instance=lambda x: isinstance(x, bool),
            str_from_instance=lambda instance: ["True" if instance else "False"],
        )

    @registry.primitive_rule
    def nonetype_rule(type_info: PrimitiveTypeInfo) -> PrimitiveConstructorSpec | None:
        if type_info.type is not type(None):
            return None
        return PrimitiveConstructorSpec(
            nargs=1,
            metavar="{None}",
            choices=("None",),
            instance_from_str=lambda args: None,
            is_instance=lambda x: x is None,
            str_from_instance=lambda instance: ["None"],
        )

    @registry.primitive_rule
    def path_rule(type_info: PrimitiveTypeInfo) -> PrimitiveConstructorSpec | None:
        try:
            if not (
                type_info.type in (os.PathLike, pathlib.Path)
                or issubclass(type_info.type, pathlib.PurePath)
            ):
                return None
        except TypeError:
            # `issubclass()` failed.
            return None
        return PrimitiveConstructorSpec(
            nargs=1,
            metavar=type_info.type.__name__.upper(),
            instance_from_str=lambda args: pathlib.Path(args[0]),
            is_instance=lambda x: hasattr(x, "__fspath__"),
            str_from_instance=lambda instance: [str(instance)],
        )

    @registry.primitive_rule
    def upath_rule(type_info: PrimitiveTypeInfo) -> PrimitiveConstructorSpec | None:
        # Check if upath is imported.
        if "upath" not in sys.modules.keys():  # pragma: no cover
            return None

        try:
            import upath

            UPath = upath.UPath
        except (ImportError, AttributeError):
            # Needed for the mock import test in
            # test_missing_optional_packages.py to pass.
            return None

        try:
            if not (type_info.type is UPath or (issubclass(type_info.type, UPath))):
                return None
        except TypeError:
            # `issubclass()` failed.
            return None

        return PrimitiveConstructorSpec(
            nargs=1,
            metavar=type_info.type.__name__.upper(),
            instance_from_str=lambda args: type_info.type(args[0]),
            is_instance=lambda x: isinstance(x, UPath),
            str_from_instance=lambda instance: [str(instance)],
        )

    @registry.primitive_rule
    def enum_rule(type_info: PrimitiveTypeInfo) -> PrimitiveConstructorSpec | None:
        if not (
            inspect.isclass(type_info.type) and issubclass(type_info.type, enum.Enum)
        ):
            return None
        cast_type = cast(Type[enum.Enum], type_info.type)
        if _markers.EnumChoicesFromValues in type_info.markers:
            choices = tuple(str(m.value) for m in cast_type)
        else:
            choices = tuple(type_info.type.__members__.keys())

        return PrimitiveConstructorSpec(
            nargs=1,
            metavar="{" + ",".join(choices) + "}",
            instance_from_str=lambda args: (
                next(
                    iter(member for member in cast_type if str(member.value) == args[0])
                )
                if _markers.EnumChoicesFromValues in type_info.markers
                else cast_type[args[0]]
            ),
            is_instance=lambda x: isinstance(x, cast_type),
            str_from_instance=lambda instance: [
                str(instance.value)
                if _markers.EnumChoicesFromValues in type_info.markers
                else instance.name
            ],
            choices=choices,
        )

    @registry.primitive_rule
    def datetime_rule(type_info: PrimitiveTypeInfo) -> PrimitiveConstructorSpec | None:
        if type_info.type not in (datetime.datetime, datetime.date, datetime.time):
            return None
        return PrimitiveConstructorSpec(
            nargs=1,
            metavar={
                datetime.datetime: "YYYY-MM-DD[THH:MM:[SS[…]]]",
                datetime.date: "YYYY-MM-DD",
                datetime.time: "HH:MM[:SS[…]]",
            }[type_info.type],
            instance_from_str=lambda args: cast(
                Union[
                    Type[datetime.datetime], Type[datetime.date], Type[datetime.time]
                ],
                type_info.type,
            ).fromisoformat(args[0]),
            is_instance=lambda x: isinstance(x, type_info.type),
            str_from_instance=lambda instance: [instance.isoformat()],
        )

    @registry.primitive_rule
    def vague_container_rule(
        type_info: PrimitiveTypeInfo,
    ) -> PrimitiveConstructorSpec | UnsupportedTypeAnnotationError | None:
        if type_info.type not in (
            dict,
            Dict,
            tuple,
            Tuple,
            list,
            List,
            collections.abc.Sequence,
            Sequence,
            set,
            Set,
        ):
            return None
        typ = type_info.type
        if typ in (dict, Dict):
            typ = Dict[str, str]
        elif typ in (tuple, Tuple):
            typ = Tuple[str, ...]  # type: ignore
        elif typ in (list, List, collections.abc.Sequence, Sequence):
            typ = List[str]
        elif typ in (set, Set):
            typ = Set[str]

        return ConstructorRegistry.get_primitive_spec(
            PrimitiveTypeInfo.make(
                typ,
                parent_markers=type_info.markers,
            )
        )

    @registry.primitive_rule
    def sequence_rule(
        type_info: PrimitiveTypeInfo,
    ) -> PrimitiveConstructorSpec | UnsupportedTypeAnnotationError | None:
        if type_info.type_origin not in (
            collections.abc.Sequence,
            frozenset,
            list,
            set,
            collections.deque,
            tuple,
        ):
            return None
        container_type = type_info.type_origin
        assert container_type is not None
        if container_type is collections.abc.Sequence:
            container_type = list

        args = get_args(type_info.type)
        if container_type is tuple:
            assert len(args) == 2
            (contained_type, ell) = args
            assert ell == Ellipsis
        elif len(args) == 1:
            (contained_type,) = args
        else:
            contained_type = Any

        inner_spec = ConstructorRegistry.get_primitive_spec(
            PrimitiveTypeInfo.make(
                raw_annotation=contained_type,
                parent_markers=type_info.markers - {_markers.UseAppendAction},
            )
        )
        if isinstance(inner_spec, UnsupportedTypeAnnotationError):
            return UnsupportedTypeAnnotationError(
                (
                    fmt.text(
                        "Could not create sequence primitive spec from ",
                        fmt.text["cyan"](f"`{type_info.type}`"),
                        " due to unsupported inner type",
                    ),
                    *inner_spec.message,
                )
            )

        # We can now handle nargs='*' with backtracking, so no need to reject it.

        def instance_from_str(args: list[str]) -> Any:
            result = parse_with_backtracking(
                args=args,
                specs=(inner_spec,),
                is_repeating=True,
            )
            if result is None:
                raise ValueError(f"Could not find valid parse for arguments: {args}")
            out = result

            assert container_type is not None
            return cast(Callable, container_type)(out)

        def str_from_instance(instance: Sequence) -> list[str]:
            out = []
            for i in instance:
                out.extend(inner_spec.str_from_instance(i))
            return out

        if _markers.UseAppendAction in type_info.markers:
            return PrimitiveConstructorSpec(
                nargs=inner_spec.nargs,
                metavar=inner_spec.metavar,
                instance_from_str=inner_spec.instance_from_str,
                is_instance=lambda x: isinstance(x, container_type)
                and all(inner_spec.is_instance(i) for i in x),
                str_from_instance=str_from_instance,
                choices=inner_spec.choices,
                _action="append",
            )
        else:
            return PrimitiveConstructorSpec(
                nargs="*",
                metavar=_strings.multi_metavar_from_single(inner_spec.metavar),
                instance_from_str=instance_from_str,
                is_instance=lambda x: isinstance(x, container_type)
                and all(inner_spec.is_instance(i) for i in x),
                str_from_instance=str_from_instance,
                choices=inner_spec.choices,
            )

    @registry.primitive_rule
    def tuple_rule(
        type_info: PrimitiveTypeInfo,
    ) -> PrimitiveConstructorSpec | UnsupportedTypeAnnotationError | None:
        if type_info.type_origin is not tuple:
            return None
        types = get_args(type_info.type)
        typeset = set(types)  # Sets are unordered.
        typeset_no_ellipsis = typeset - {Ellipsis}  # type: ignore

        if typeset_no_ellipsis != typeset:
            # Ellipsis: variable argument counts. When an ellipsis is used, tuples must
            # contain only one type.
            assert len(typeset_no_ellipsis) == 1
            return sequence_rule(type_info)

        inner_specs: list[PrimitiveConstructorSpec] = []
        for contained_type in types:
            spec = ConstructorRegistry.get_primitive_spec(
                PrimitiveTypeInfo.make(contained_type, type_info.markers)
            )
            if isinstance(spec, UnsupportedTypeAnnotationError):
                return UnsupportedTypeAnnotationError(
                    (
                        fmt.text(
                            "Could not create tuple primitive spec from ",
                            fmt.text["cyan"](f"`{type_info.type}`"),
                            " due to unsupported inner type",
                        ),
                        *spec.message,
                    )
                )
            inner_specs.append(spec)

        def instance_from_str(args: list[str]) -> tuple:
            # Use backtracking for all cases (both fixed and variable nargs).
            # Complexity is bad, O(k^n), where k is the max number of nargs.
            # options and n is the number of tuple elements. We could revisit,
            # but in practice k and n should both be small.
            result = parse_with_backtracking(
                args=args,
                specs=tuple(inner_specs),
                is_repeating=False,
            )
            if result is None:
                raise ValueError(
                    f"Could not parse arguments {args} into tuple type {type_info.type}"
                )
            return tuple(result)

        def str_from_instance(instance: tuple) -> list[str]:
            out = []
            for member, spec in zip(instance, inner_specs):
                out.extend(spec.str_from_instance(member))
            return out

        # Compute all possible total argument counts.
        nargs = _compute_total_nargs(inner_specs)

        return PrimitiveConstructorSpec(
            nargs=nargs,
            metavar=" ".join(spec.metavar for spec in inner_specs),
            instance_from_str=instance_from_str,
            str_from_instance=str_from_instance,
            is_instance=lambda x: isinstance(x, tuple)
            and len(x) == len(inner_specs)
            and all(spec.is_instance(member) for member, spec in zip(x, inner_specs)),
        )

    @registry.primitive_rule
    def dict_rule(
        type_info: PrimitiveTypeInfo,
    ) -> PrimitiveConstructorSpec | UnsupportedTypeAnnotationError | None:
        if (
            type_info.type_origin not in (dict, collections.abc.Mapping)
            or len(get_args(type_info.type)) != 2
        ):
            return None

        key_type, val_type = get_args(type_info.type)
        key_spec = ConstructorRegistry.get_primitive_spec(
            PrimitiveTypeInfo.make(
                raw_annotation=key_type,
                parent_markers=type_info.markers,
            )
        )
        val_spec = ConstructorRegistry.get_primitive_spec(
            PrimitiveTypeInfo.make(
                raw_annotation=val_type,
                parent_markers=type_info.markers - {_markers.UseAppendAction},
            )
        )
        if isinstance(key_spec, UnsupportedTypeAnnotationError):
            return UnsupportedTypeAnnotationError(
                (
                    fmt.text(
                        "Could not create dict primitive spec from ",
                        fmt.text["cyan"](f"`{type_info.type}`"),
                        " due to unsupported key type",
                    ),
                    *key_spec.message,
                )
            )
        if isinstance(val_spec, UnsupportedTypeAnnotationError):
            return UnsupportedTypeAnnotationError(
                (
                    fmt.text(
                        "Could not create dict primitive spec from ",
                        fmt.text["cyan"](f"`{type_info.type}`"),
                        " due to unsupported value type",
                    ),
                    *val_spec.message,
                )
            )
        pair_metavar = f"{key_spec.metavar} {val_spec.metavar}"

        def instance_from_str(args: list[str]) -> dict:
            out = {}

            # For UseAppendAction, we need to determine if we're parsing:
            # 1. A single key-value pair (when nargs is fixed).
            # 2. Multiple key-value pairs (when nargs is '*').
            if _markers.UseAppendAction in type_info.markers:
                # Check if we have a fixed number of args for a single pair.
                if isinstance(key_spec.nargs, int) and isinstance(val_spec.nargs, int):
                    # Fixed size: parse single key-value pair.
                    is_repeating = False
                else:
                    # Variable size: parse multiple pairs.
                    is_repeating = True
            else:
                # Without UseAppendAction, always parse multiple pairs.
                is_repeating = True

            parsed = parse_with_backtracking(
                args, (key_spec, val_spec), is_repeating=is_repeating
            )
            if parsed is None:
                raise ValueError("Failed to parse key-value pairs!")

            # When is_repeating=True, parse_with_backtracking alternates between
            # key_spec and val_spec, so parsed contains [key1, val1, key2, val2, ...].
            for i in range(0, len(parsed), 2):
                out[parsed[i]] = parsed[i + 1]

            return out

        def str_from_instance(instance: dict) -> list[str]:
            # TODO: this may be strange right now for the append action.
            out: list[str] = []
            for key, value in instance.items():
                out.extend(key_spec.str_from_instance(key))
                out.extend(val_spec.str_from_instance(value))
            return out

        if _markers.UseAppendAction in type_info.markers:
            # Compute all possible total argument counts for dict key-value pairs.
            nargs = _compute_total_nargs([key_spec, val_spec])

            return PrimitiveConstructorSpec(
                nargs=nargs,
                metavar=pair_metavar,
                instance_from_str=instance_from_str,
                is_instance=lambda x: isinstance(x, dict)
                and all(
                    key_spec.is_instance(k) and val_spec.is_instance(v)
                    for k, v in x.items()
                ),
                str_from_instance=str_from_instance,
                _action="append",
            )
        else:
            return PrimitiveConstructorSpec(
                nargs="*",
                metavar=_strings.multi_metavar_from_single(pair_metavar),
                instance_from_str=instance_from_str,
                is_instance=lambda x: isinstance(x, dict)
                and all(
                    key_spec.is_instance(k) and val_spec.is_instance(v)
                    for k, v in x.items()
                ),
                str_from_instance=str_from_instance,
            )

    @registry.primitive_rule
    def literal_rule(type_info: PrimitiveTypeInfo) -> PrimitiveConstructorSpec | None:
        if not is_typing_literal(type_info.type_origin):
            return None
        choices = get_args(type_info.type)
        str_choices = tuple(
            (
                (
                    x.value
                    if _markers.EnumChoicesFromValues in type_info.markers
                    else x.name
                )
                if isinstance(x, enum.Enum)
                else str(x)
            )
            for x in choices
        )
        return PrimitiveConstructorSpec(
            nargs=1,
            metavar="{" + ",".join(str_choices) + "}",
            instance_from_str=lambda args: choices[str_choices.index(args[0])],
            is_instance=lambda x: x in choices,
            str_from_instance=lambda instance: [str(instance)],
            choices=str_choices,
        )

    @registry.primitive_rule
    def union_rule(
        type_info: PrimitiveTypeInfo,
    ) -> PrimitiveConstructorSpec | UnsupportedTypeAnnotationError | None:
        if not is_typing_union(type_info.type_origin):
            return None
        options = list(get_args(type_info.type))
        if type(None) in options:
            # Move `None` types to the beginning.
            # If we have `Optional[str]`, we want this to be parsed as
            # `Union[NoneType, str]`.
            options.remove(type(None))
            options.insert(0, type(None))

        # General unions, eg Union[int, bool]. We'll try to convert these from left to
        # right.
        option_specs: dict[TypeForm[object], PrimitiveConstructorSpec] = {}
        choices: tuple[str, ...] | None = ()
        nargs: int | tuple[int, ...] | Literal["*"] = 1
        first = True
        all_fixed_nargs = True
        nargs_set: set[int] = set()

        metavar_parts: list[str] = []

        for t in options:
            option_type_info = PrimitiveTypeInfo.make(
                raw_annotation=t,
                parent_markers=type_info.markers,
            )

            # If the argument is not suppressed, we can add the ability to
            # suppress individual options.
            if (
                _markers.Suppress not in type_info.markers
                and _markers.Suppress in option_type_info.markers
            ):
                continue

            option_spec = ConstructorRegistry.get_primitive_spec(option_type_info)
            if isinstance(option_spec, UnsupportedTypeAnnotationError):
                return UnsupportedTypeAnnotationError(
                    (
                        fmt.text(
                            "Could not create union primitive spec from ",
                            fmt.text["cyan"](f"`{type_info.type}`"),
                            " due to unsupported option type",
                        ),
                        *option_spec.message,
                    )
                )
            if option_spec.choices is None:
                choices = None
            elif choices is not None:
                choices = choices + option_spec.choices

            option_specs[t] = option_spec
            if not (t is type(None) and _markers.DisallowNone in type_info.markers):
                metavar_parts.append(option_spec.metavar)

            if t is not type(None):
                # Track if all options have fixed nargs.
                if isinstance(option_spec.nargs, int):
                    nargs_set.add(option_spec.nargs)
                else:
                    all_fixed_nargs = False

                # Enforce that `nargs` is the same for all child types, except for
                # NoneType.
                if first:
                    nargs = option_spec.nargs
                    first = False
                elif nargs != option_spec.nargs:
                    # If all options have fixed nargs (even if different), collect them.
                    if all_fixed_nargs and isinstance(option_spec.nargs, int):
                        continue  # We'll handle this after the loop.
                    else:
                        # Just be as general as possible if we see inconsistencies.
                        nargs = "*"

        # If we have multiple fixed nargs values, use a tuple.
        if all_fixed_nargs and len(nargs_set) > 1:
            nargs = tuple(sorted(nargs_set))

        metavar: str
        metavar = _strings.join_union_metavars(metavar_parts)

        def union_instantiator(strings: list[str]) -> Any:
            errors = []
            for i, (t, option_spec) in enumerate(option_specs.items()):
                # Check if NoneType is disallowed.
                if t is type(None) and _markers.DisallowNone in type_info.markers:
                    continue

                # Check choices.
                if option_spec.choices is not None and any(
                    x not in option_spec.choices for x in strings
                ):
                    errors.append(
                        f"{options[i]}: {strings} does not match choices {option_spec.choices}"
                    )
                    continue

                # Try passing input into instantiator.
                nargs_match = False
                if isinstance(option_spec.nargs, int):
                    nargs_match = len(strings) == option_spec.nargs
                elif isinstance(option_spec.nargs, tuple):
                    nargs_match = len(strings) in option_spec.nargs
                elif option_spec.nargs == "*":
                    nargs_match = True

                if nargs_match:
                    try:
                        return option_spec.instance_from_str(strings)
                    except ValueError as e:
                        # Failed, try next instantiator.
                        errors.append(f"{options[i]}: {e.args[0]}")
                else:
                    errors.append(
                        f"{options[i]}: input length {len(strings)} did not match expected"
                        f" argument count {option_spec.nargs} of `[bold]{option_spec.metavar}[/bold]`"
                    )
            raise ValueError(
                f"no type matching {metavar} could be instantiated from"
                f" {strings}.\n\nGot errors:  \n- " + "\n- ".join(errors)
            )

        def str_from_instance(instance: Any) -> list[str]:
            fuzzy_match = None
            for option_spec in option_specs.values():
                is_instance = option_spec.is_instance(instance)
                if is_instance is True:
                    return option_spec.str_from_instance(instance)
                elif is_instance == "~":
                    fuzzy_match = option_spec

            # If we get here, we have a fuzzy match.
            if fuzzy_match is not None:
                return fuzzy_match.str_from_instance(instance)

            assert False, (
                f"could not match default value {instance} with any types in union {options}"
            )

        return PrimitiveConstructorSpec(
            nargs=nargs,
            metavar=metavar,
            instance_from_str=union_instantiator,
            is_instance=lambda x: any(
                spec.is_instance(x) for spec in option_specs.values()
            ),
            str_from_instance=str_from_instance,
            choices=None if choices is None else tuple(set(choices)),
        )

    @registry.primitive_rule
    def python_syntax_collections_rule(
        type_info: PrimitiveTypeInfo,
    ) -> PrimitiveConstructorSpec | UnsupportedTypeAnnotationError | None:
        """Handle collections with Python literal syntax when UsePythonSyntaxForLiteralCollections marker is present."""
        # Check if the marker is present.
        if _markers.UsePythonSyntaxForLiteralCollections not in type_info.markers:
            return None

        # Check if the type is a collection type.
        if type_info.type_origin not in (list, tuple, set, dict, frozenset):
            return None

        # Return None if no type arguments.
        type_args = get_args(type_info.type)
        if len(type_args) == 0:
            return None

        # Validate that all innermost types are compatible with ast.literal_eval.
        def is_literal_eval_compatible(typ: Any) -> bool:
            """Check if a type is compatible with ast.literal_eval().

            ast.literal_eval() only supports: str, bytes, int, float, complex, bool, None,
            and the collection types list, tuple, dict, set. typing.Literal types are also
            supported. Nested structures of these types are supported.
            """
            from typing import Literal as LiteralType

            # Ellipsis is used in variable-length tuples like tuple[int, ...].
            if typ is Ellipsis:
                return True

            origin = get_origin(typ)

            # Handle Literal types - check if all values are literal-eval compatible.
            if origin is LiteralType:
                # Literal values themselves must be literal-eval compatible types.
                return all(
                    type(arg) in (str, bytes, int, float, complex, bool, type(None))
                    for arg in get_args(typ)
                )

            if origin is not None:
                # Recurse into generic types.
                return all(is_literal_eval_compatible(arg) for arg in get_args(typ))

            # Whitelist only types that ast.literal_eval() actually supports.
            # This excludes built-in types like frozenset, range, slice, etc.
            return typ in (
                str,
                bytes,
                int,
                float,
                complex,
                bool,
                type(None),
                list,
                tuple,
                dict,
                set,
            )

        # Return None if any types are incompatible - let other rules handle it.
        if not all(is_literal_eval_compatible(arg) for arg in type_args):
            return None

        def instance_from_str(args: list[str]) -> Any:
            import ast

            try:
                # Use ast.literal_eval() for safe parsing of Python literals.
                # Only supports: strings, bytes, numbers, tuples, lists, dicts, sets,
                # booleans, None, and nested structures of these types.
                value = ast.literal_eval(args[0])
            except (ValueError, SyntaxError) as e:
                raise ValueError(
                    f"Could not parse '{args[0]}' as Python literal: {e}"
                ) from e

            # Validate the type using typeguard.
            if not _resolver.is_instance(type_info.type, value):
                raise ValueError(
                    f"Value {value} (type {type(value)}) does not match expected type {type_info.type}"
                )
            return value

        def str_from_instance(instance: Any) -> list[str]:
            return [repr(instance)]

        def is_instance_fn(x: Any) -> bool:
            return _resolver.is_instance(type_info.type, x)

        # Build a nice metavar showing the structure.
        def get_metavar_for_type(typ: Any) -> str:
            """Recursively build metavar string for a type."""
            from typing import Literal as LiteralType

            origin = get_origin(typ)

            # Handle Literal types - show the literal values.
            if origin is LiteralType:
                args = get_args(typ)
                # Format as {val1,val2,val3}.
                return "{" + ",".join(repr(arg) for arg in args) + "}"

            if origin is None:
                # Handle None type specially - show as "None" instead of "NONETYPE".
                if typ is type(None):
                    return "None"
                # Primitive type - just use uppercase name.
                return typ.__name__.upper()

            # Get type arguments.
            args = get_args(typ)

            if origin in (list, List):
                elem = get_metavar_for_type(args[0])
                return f"[{elem},...]"
            elif origin in (tuple, Tuple):
                if len(args) == 2 and args[1] is Ellipsis:
                    # Variable-length tuple.
                    elem = get_metavar_for_type(args[0])
                    return f"({elem},...)"
                else:
                    # Fixed-length tuple.
                    elems = ",".join(get_metavar_for_type(arg) for arg in args)
                    return f"({elems})"
            elif origin in (set, Set):
                elem = get_metavar_for_type(args[0])
                return f"{{{elem},...}}"
            elif origin in (dict, Dict):
                key = get_metavar_for_type(args[0])
                val = get_metavar_for_type(args[1])
                return f"{{{key}:{val},...}}"
            else:
                return str(origin).upper()

        metavar = get_metavar_for_type(type_info.type)

        return PrimitiveConstructorSpec(
            nargs=1,
            metavar=metavar,
            instance_from_str=instance_from_str,
            str_from_instance=str_from_instance,
            is_instance=is_instance_fn,
        )



from __future__ import annotations

from typing import Annotated, Generic, TypeVar

from pydantic import BaseModel

import tyro


class AThing(BaseModel):
    a: int


TContainsAThing = TypeVar("TContainsAThing", bound=AThing)


class GenericThingManager(Generic[TContainsAThing]):
    @classmethod
    def spam(
        cls,
        thing2: TContainsAThing,
        thing: Annotated[
            TContainsAThing | list[TContainsAThing],
            tyro.conf.arg(constructor_factory=lambda: TContainsAThing),  # type: ignore
        ],
    ) -> tuple[TContainsAThing, TContainsAThing]:
        return thing2, thing  # type: ignore

    @classmethod
    def eggs(
        cls,
        thing: Annotated[
            TContainsAThing | list[TContainsAThing],
            tyro.conf.arg(constructor_factory=lambda: TContainsAThing),  # type: ignore
        ],
    ) -> TContainsAThing:
        return thing  # type: ignore


class ABThing(AThing):
    b: str


class ABThingManager(GenericThingManager[ABThing]):
    pass


def test_generic_constructor_factory_spam():
    """Test that spam subcommand works with generic constructor factory."""
    # Test with single ABThing instance.
    thing2, thing = tyro.cli(
        ABThingManager.spam,
        args=[
            "--thing2.a",
            "1",
            "--thing2.b",
            "foo",
            "--thing.a",
            "2",
            "--thing.b",
            "bar",
        ],
    )
    assert isinstance(thing2, ABThing)
    assert thing2.a == 1
    assert thing2.b == "foo"
    assert isinstance(thing, ABThing)
    assert thing.a == 2
    assert thing.b == "bar"


def test_generic_constructor_factory_eggs():
    """Test that eggs subcommand works with generic constructor factory."""
    # Test with single ABThing instance.
    thing = tyro.cli(
        ABThingManager.eggs,
        args=["--thing.a", "3", "--thing.b", "baz"],
    )
    assert isinstance(thing, ABThing)
    assert thing.a == 3
    assert thing.b == "baz"


def test_generic_constructor_factory_subcommand_cli():
    """Test subcommand CLI with generic constructor factory."""
    # Test spam subcommand.
    thing2, thing = tyro.extras.subcommand_cli_from_dict(
        {
            "spam": ABThingManager.spam,
            "eggs": ABThingManager.eggs,
        },
        args=[
            "spam",
            "--thing2.a",
            "1",
            "--thing2.b",
            "hello",
            "--thing.a",
            "2",
            "--thing.b",
            "world",
        ],
    )
    assert isinstance(thing2, ABThing)
    assert thing2.a == 1
    assert thing2.b == "hello"
    assert isinstance(thing, ABThing)
    assert thing.a == 2
    assert thing.b == "world"

    # Test eggs subcommand.
    thing = tyro.extras.subcommand_cli_from_dict(
        {
            "spam": ABThingManager.spam,
            "eggs": ABThingManager.eggs,
        },
        args=["eggs", "--thing.a", "4", "--thing.b", "test"],
    )
    assert isinstance(thing, ABThing)
    assert thing.a == 4
    assert thing.b == "test"
